#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "crcmodel.h"
#include "Processor.h"

static unsigned char lastSeq;

_sProcessorInfo *ProcessorInit( int showAllData )
	{
	_sProcessorInfo *retval = (_sProcessorInfo *)malloc( sizeof(_sProcessorInfo) );
	if ( retval != NULL )
		{
		retval->state = state_WaitPacketType;
		retval->packet.cpPayload = NULL;
		retval->payloadOffset = 0;
		retval->escReceived = 0;
		retval->showAllData = showAllData;
		}

	lastSeq = 0xFF;

	return retval;
	}

void ProcessorCleanup( _sProcessorInfo *info )
	{
	if ( info != NULL )
		{
		// free any previously-allocated payload buffer
		if ( info->packet.cpPayload != NULL )
			{
			free( info->packet.cpPayload );
			info->packet.cpPayload = NULL;
			}

		free( info );
		}
	}

void ProcessByte( _sProcessorInfo *info, unsigned char ch )
	{
	// This code processes individual bytes, running a little state machine
	// to decode complete packets.
	// It expects "complete" packets - i.e. it's not currently set up to handle
	// start-of-packet (x10 x78) or end-of-packet (x10 x03) indications within
	// a partially-decoded packet.

	if ( ch == 0x10 )
		{
		if ( !info->escReceived )
			{
			info->escReceived = true;
			return;
			}
		}

	switch ( info->state )
		{
		default:
			info->state = state_WaitPacketType;

		case state_WaitPacketType:
			if ( info->escReceived )
				{
				info->packet.cPacketType = ch;
				info->state = state_WaitLenHi;
				}
			break;

		case state_WaitLenHi:
			info->packet.iLen = (ch << 8);
			info->state = state_WaitLenLo;
			break;

		case state_WaitLenLo:
			info->packet.iLen |= ch;
			info->state = state_WaitSeq;
			break;

		case state_WaitSeq:
			info->packet.cSeq = ch;
			info->state = state_WaitID;
			break;

		case state_WaitID:
			info->packet.cID = ch;
			// free any previously-allocated payload buffer
			if ( info->packet.cpPayload != NULL )
				{
				free( info->packet.cpPayload );
				info->packet.cpPayload = NULL;
				}
			if ( info->packet.iLen - 2 )
				{
				info->packet.cpPayload = (unsigned char *)malloc( info->packet.iLen - 2 );
				info->payloadOffset = 0;
				info->state = state_WaitPayload;
				}
			else
				{
				info->state = state_WaitCRCHi;
				}
			break;

		case state_WaitPayload:
			info->packet.cpPayload[ info->payloadOffset ] = ch;
			if ( ++info->payloadOffset == info->packet.iLen-2 )
				{
				info->state = state_WaitCRCHi;
				}
			break;

		case state_WaitCRCHi:
			info->packet.iCRC = (ch << 8);
			info->state = state_WaitCRCLo;
			break;

		case state_WaitCRCLo:
			info->packet.iCRC |= ch;
			info->state = state_WaitTerminator;
			break;

		case state_WaitTerminator:
			if ( info->escReceived )
				{
				cm_t cm;					// a "CRC Model" structure
				
				cm.cm_width = 16;			// 16-bit
				cm.cm_poly = 0x1021;		// CCITT-16 polynomial (X16  X12  X5  X0)
				cm.cm_init = 0x0000;		// seed value 0x0000
				cm.cm_refin = TRUE;			// input data reflected (so it's lsb-first)
				cm.cm_refot = TRUE;			// output CRC reflected
				cm.cm_xorot = 0x0000;		// XOR resulting CRC w/0x0000 (i.e. nothing)

				cm_ini( &cm );				// initialize the CRC Model (just sets seed)
				
				// the CRC is calculated on the entire packet, not including the 0x10
				// 'escape' bytes
				cm_nxt( &cm, info->packet.cPacketType );	// packet type
				cm_nxt( &cm, info->packet.iLen >> 8 );		// len MSB
				cm_nxt( &cm, info->packet.iLen & 0xFF );	// len LSB
				cm_nxt( &cm, info->packet.cSeq );			// seq #
				cm_nxt( &cm, info->packet.cID );			// 'ID'
				cm_blk( &cm, info->packet.cpPayload, info->packet.iLen-2 );	// payload bytes

				unsigned short crc = (unsigned short) cm_crc( &cm );	// get the resulting CRC
				if ( crc == info->packet.iCRC )
					{
					ProcessPacket( info );
					}
				else
					{
					fprintf( stdout, "BAD CRC. Expected %04X, recd %04X\n\n",
						crc, info->packet.iCRC );
					}
				}

			info->state = state_WaitPacketType;
			break;
		}

	info->escReceived = false;
	}

void ProcessPacket( _sProcessorInfo *info )
	{
	_sPacket *p = &info->packet;

	fprintf( stdout, "\n" );

	if ( info->showAllData )
		{
		IndentedPrintf( p->cID, stdout, "%s, from %s\n",
			p->cPacketType == 0x70 ? "DELPHI   " :
			p->cPacketType == 0x78 ? "DELPHI_EX" :
									 "?????????",
			p->cID == 0x04 ? "DCT" : "\"Sidecar\"" );
		}

	unsigned char seq = p->cSeq & 0x33;

	IndentedPrintf( p->cID, stdout, "" );

	// this code will probably print out a couple of "but the xxx nybble is wrong"
	// messages when the first packets are received. The sequence numbers don't
	// seem to match what's expected when the DCT first powers up.
	if ( lastSeq != 0xFF )
		{
		if ( p->cSeq & 0x80 )
			{
			// we're assuming that a sequence number with the high bit set is
			// an ACK for the previous packet. see if the other bits look right.
			fprintf( stdout, "ACK " );
			if ( seq == ((lastSeq + 0x01) & 0x33) )
				{
//				fprintf( stdout, "ACK " );
				}
			else
			if ( info->showAllData )
				{
				fprintf( stdout,
					", but low nybble wrong (got %02X, exp %02X) ",
					seq, ((lastSeq + 0x01) & 0x33) );
				}
			}
		else
		if ( info->showAllData )
			{
			if ( seq == ((lastSeq + 0x10) & 0x33) )
				{
//				fprintf( stdout, ", new packet" );
				}
			else
				{
				fprintf( stdout,
						"[high nybble wrong (got %02X, exp %02X)] ",
						seq, ((lastSeq + 0x10) & 0x33) );
				}
			}
		}

	lastSeq = seq;

	if ( p->cPacketType == 0x70 )
		{
		fprintf( stdout, "Startup msg\n" );
		}
	else
	if ( p->cPacketType == 0x78 )
		{
		if ( p->iLen > 2 )
			{
			if ( info->showAllData )
				{
				fprintf( stdout, "cmd=%02X\n", p->cpPayload[0] );
				}
			switch( p->cpPayload[0] )
				{
				default:
					fprintf( stdout, "unknown packet type\n" );
					break;

				case _PT_ChannelStatusReq:
					fprintf( stdout, "ChannelStatusReq or 'Go DELPHI_EX' (?)\n" );
					break;

				case _PT_ChannelStatus:
					{
					int channel =  (p->cpPayload[1] << 8) | p->cpPayload[2];

					if ( info->showAllData )
						{
						fprintf( stdout, "ChannelStatus: " );
						}

					fprintf( stdout, "CH:%4d ", channel );

					int type = p->cpPayload[3];
					switch ( type )
						{
						default:	// ???
							fprintf( stdout, "[unknown type]\n" );
							break;

						case 0:		// analog
							fprintf( stdout, "[ANALOG]\n" );
							break;

						case 1:		// digital
							fprintf( stdout, "[DIGITAL]\n" );

							if ( info->showAllData )
								{
								time_t t = (p->cpPayload[6] << 24) |
										   (p->cpPayload[7] << 16) |
										   (p->cpPayload[8] <<  8) |
										   (p->cpPayload[9] <<  0);
								// add offset for GPS-UTC time
								time_t tadj = t + 0x12D53D80;

								struct tm *gmt;
								gmt = gmtime( &tadj );
								char *tstr = asctime( gmt );
								tstr[strlen(tstr)-1] = 0;
								IndentedPrintf( p->cID, stdout, "    epoch time: %s\n", tstr );

								time_t tp = time( NULL );;
								tstr = ctime( &tp );
								tstr[strlen(tstr)-1] = 0;
								IndentedPrintf( p->cID, stdout, "       PC time: %s\n", tstr );

								IndentedPrintf( p->cID, stdout, "       DCT:%08X\n", t );
								IndentedPrintf( p->cID, stdout, "    DCTadj:%08X\n", tadj );
								IndentedPrintf( p->cID, stdout, "        PC:%08X\n", tp );
								}
							break;
						}
					}
					break;

				case _PT_IRKeypress:
					if ( info->showAllData )
						{
						fprintf( stdout, "IR data: %02X, key: ",
								p->cpPayload[1] );
						}
					fprintf( stdout, "%s\n",
								GetKeyName( p->cpPayload[1]) );
					break;
				}
			}
		else	// no payload
			{
			fprintf( stdout, "\n" );
			}
		}
	else
		// neither 0x70 nor 0x78
		{
		fprintf( stdout, "<unknown message type>\n" );
		}

	if ( info->showAllData )
		{
		fprintf( stdout, "\n" );
		}
	}


// The GI remote protocol specifies that remote controls send codes
// between 0x00 and 0x3F, inclusive...
#define _NumKeyNames	0x40
const char *KeyNames[ _NumKeyNames ] =
	{
	/* 0x00 */  "0",
	/* 0x01 */  "1",
	/* 0x02 */  "2",
	/* 0x03 */  "3",
	/* 0x04 */  "4",
	/* 0x05 */  "5",
	/* 0x06 */  "6",
	/* 0x07 */  "7",
	/* 0x08 */  "8",
	/* 0x09 */  "9",
	/* 0x0A */  "POWER",
	/* 0x0B */  "CH+",
	/* 0x0C */  "CH-",
	/* 0x0D */  "VOL+",
	/* 0x0E */  "VOL-",
	/* 0x0F */  "MUTE",
	/* 0x10 */  "MUSIC",
	/* 0x11 */  "SELECT/OK/ENTER",
	/* 0x12 */  "EXIT",
	/* 0x13 */  "LAST",
	/* 0x14 */  "SWITCH_AB",
	/* 0x15 */  "FAVORITE",
	/* 0x16 */  "PPV",
	/* 0x17 */  "A",
	/* 0x18 */  "P_F",
	/* 0x19 */  "MENU",
	/* 0x1A */  "<unknown>",
	/* 0x1B */  "<unknown>",
	/* 0x1C */  "OUTPUT_CHAN",
	/* 0x1D */  "FAST_FORWARD",
	/* 0x1E */  "REVERSE",
	/* 0x1F */  "PAUSE",
	/* 0x20 */  "<unknown>",
	/* 0x21 */  "<unknown>",
	/* 0x22 */  "BROWSE",
	/* 0x23 */  "CANCEL",
	/* 0x24 */  "LOCK",
	/* 0x25 */  "LIST",
	/* 0x26 */  "THEME",
	/* 0x27 */  "B",
	/* 0x28 */  "C",
	/* 0x29 */  "<unknown>",
	/* 0x2A */  "<unknown>",
	/* 0x2B */  "<unknown>",
	/* 0x2C */  "<unknown>",
	/* 0x2D */  "<unknown>",
	/* 0x2E */  "<unknown>",
	/* 0x2F */  "<unknown>",
	/* 0x30 */  "GUIDE",
	/* 0x31 */  "RECORD",
	/* 0x32 */  "HELP",
	/* 0x33 */  "INFO",
	/* 0x34 */  "UP",
	/* 0x35 */  "DOWN",
	/* 0x36 */  "LEFT",
	/* 0x37 */  "RIGHT",
	/* 0x38 */  "DAY+",
	/* 0x39 */  "DAY-",
	/* 0x3A */  "PAGE+",
	/* 0x3B */  "PAGE-",
	/* 0x3C */  "<unknown>",
	/* 0x3D */  "<unknown>",
	/* 0x3E */  "<unknown>",
	/* 0x3F */  "<unknown>"
	};

const char *GetKeyName( unsigned char key )
	{
	return key < _NumKeyNames ? KeyNames[ key ] : "<invalid>";
	}


void SenderIndent( unsigned char who, FILE *stream )
	{
	// DCT messages on the right side, Sidecar on the left
	if ( who == _SID_DCT )
		{
		fprintf( stream, "%40.40s", " " );
		}
	}
void IndentedPrintf( unsigned char who, FILE *stream, char *fmt, ... )
	{
	SenderIndent( who, stream );
	va_list args;
	va_start( args, fmt );
	vfprintf( stream, fmt, args );
	va_end( args );
	}
