#include <stdio.h>
#include <stdlib.h>

#include "serial.h"

void ShowError( char *fmt, ... )
	{
	va_list args;
	va_start( args, fmt );
	vfprintf( stderr, fmt, args );
	va_end( args );

#ifdef _WIN32
	fprintf( stderr,"  errno==0x%08X\n", ::GetLastError() );
#else
	fprintf( stderr, "\nNEED PLATFORM-SPECIFIC ERROR DISPLAY CODE\n" );
#endif
	}

_ser_struct *ser_open(  int port,
                    int bps,
                    int data,
                    _eSer_Stop stop,
                    _eSer_Parity parity )
	{
	_ser_struct *retval = NULL;
	
#ifdef _WIN32
	_ser_struct *s = (_ser_struct *)malloc( sizeof(_ser_struct) );
	if ( s != NULL )
		{
		memset( s, 0, sizeof(_ser_struct) );

		// max COM port number is 256 (should probably check for that),
		// so the string can only be 11 chars long
		char portName[16];
		sprintf( portName, "\\\\.\\COM%d", port+1 );
		if ( (s->handle = ::CreateFile( portName,
							GENERIC_READ | GENERIC_WRITE,
							0,
							0,
							OPEN_EXISTING,
							FILE_FLAG_OVERLAPPED,
							0))
			== INVALID_HANDLE_VALUE )
			{
			ShowError( "ser_open() - CreateFile(%s)", portName );
			}
		else
		if ( (s->overlapped.hEvent = ::CreateEvent( NULL, TRUE, FALSE, NULL ))
			== NULL )
			{
			ShowError( "ser_open() - CreateEvent() failed" );
			}
		else
		if ( ! ser_configure( s, bps, data, stop, parity ) )
			{
			ShowError( "ser_open() - ser_configure() failed" );
			}
		else
			{
			retval = s;
			}
		if ( retval == NULL )
			{
			// close the port!
			if ( s->handle )
				{
				::CloseHandle( s->handle );
				}
			// close the event!
			if ( s->overlapped.hEvent )
				{
				::CloseHandle( s->overlapped.hEvent );
				}
			free( s );
			}
		}
#endif	// _WIN32

	return retval;
	}

void ser_close( _ser_struct *port )
	{
	if ( port != NULL && port->handle != NULL )
		{
#ifdef _WIN32
		::CloseHandle( port->overlapped.hEvent );
		::CloseHandle( port->handle );
		free( port );
#endif	// _WIN32
		}
	}

int ser_configure(  _ser_struct *port,
                    int bps,
                    int data,
                    _eSer_Stop stop,
                    _eSer_Parity parity )
	{
	if ( port == NULL )
		{
		fprintf( stdout, "ser_configure() - param is NULL!\n" );
		return 0;
		}

#ifdef _WIN32
	DCB dcb;

	if ( ! ::GetCommState( port->handle, &dcb ) )
		{
		ShowError( "ser_configure() - GetCommState()" );
		return 0;
		}

	dcb.BaudRate = bps;
	dcb.fBinary = 1;
	dcb.fParity = (parity == _ser_parity_none ? 0 : 1);
	dcb.fOutxCtsFlow = 0;
	dcb.fOutxDsrFlow = 0;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	dcb.fDsrSensitivity = 0;
	dcb.fTXContinueOnXoff = 1;
	dcb.fOutX = 0;
	dcb.fInX = 0;
	dcb.fErrorChar = 0;
	dcb.fNull = 0;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	dcb.fAbortOnError = 0;
	dcb.wReserved = 0;
	dcb.XonLim = 0;
	dcb.XoffLim = 0;
	dcb.ByteSize = data;
	if ( parity == _ser_parity_even )
		{
		dcb.Parity = EVENPARITY;
		}
	else if ( parity == _ser_parity_odd )
		{
		dcb.Parity = ODDPARITY;
		}
	else
		{
		dcb.Parity = NOPARITY;
		}
	if ( stop == _ser_stop_1 )
		{
		dcb.StopBits = ONESTOPBIT;
		}
	else if ( stop == _ser_stop_2 )
		{
		dcb.StopBits = TWOSTOPBITS;
		}
	else
		{
		dcb.StopBits = ONE5STOPBITS;
		}
	dcb.XonChar = 0;
	dcb.XoffChar = 1;
	dcb.ErrorChar = 0;
	dcb.EofChar = 0;
	dcb.EvtChar = 0;

	if ( ! ::SetCommState( port->handle, &dcb ) )
		{
		ShowError( "ser_configure() - SetCommState()" );
		return 0;
		}

	int wordBits = 1 + data + (stop == _ser_stop_1 ? 1 : 2);
	if ( parity != _ser_parity_none )
		wordBits++;
	port->byteTime = (double)1000.0 / ((double)bps / (double)wordBits);

	COMMTIMEOUTS to;
	to.ReadIntervalTimeout = 0;
	to.ReadTotalTimeoutConstant = 0;
	to.ReadTotalTimeoutMultiplier = 0;
	to.WriteTotalTimeoutConstant = 0;
	to.WriteTotalTimeoutMultiplier = 0;
	if ( ! ::SetCommTimeouts( port->handle, &to ) )
		{
		ShowError( "ser_configure() - SetCommTimeouts()" );
		return 0;
		}	
#endif	// _WIN32

	return 1;
	}

int ser_peek( _ser_struct *port )
	{
	if ( port == NULL )
		{
		fprintf( stdout, "ser_peek() - param is NULL!\n" );
		return 0;
		}

#ifdef _WIN32
	COMSTAT comstat;
	DWORD dwErrorMask;
	::ClearCommError( port->handle, &dwErrorMask, &comstat );
	return comstat.cbInQue;
#endif	// _WIN32

	return 0;
	}

int ser_read(   _ser_struct *port,
                unsigned char *buffer,
                int count,
                int timeout )
	{
	if ( port == NULL )
		{
		fprintf( stdout, "ser_read() - param is NULL!\n" );
		return 0;
		}

#ifdef _WIN32
	DWORD dwBytesRead = 0;

	port->overlapped.Offset = 0;
	port->overlapped.OffsetHigh = 0;

	COMSTAT comstat;
	DWORD dwErrorMask;
	::ClearCommError( port->handle, &dwErrorMask, &comstat );

	if ( dwErrorMask )
		{
		fprintf( stderr,
				"ser_read() - ClearCommError returns 0x%08X\n",
				dwErrorMask );
		}

	if ( ! ::ReadFile( port->handle, buffer, count, &dwBytesRead, &port->overlapped ) )
		{
		DWORD error = ::GetLastError();
		if ( error != ERROR_IO_PENDING )
			{
			ShowError( "ser_read() - ReadFile()" );
			}
		else
			{
			switch ( ::WaitForSingleObject( port->overlapped.hEvent, (DWORD)timeout ) )
				{
				case WAIT_OBJECT_0:
					if ( ! ::GetOverlappedResult( port->handle, &port->overlapped, &dwBytesRead, FALSE ) )
						{
						ShowError( "ser_read() - GetOverlappedResult()" );
						}
					::ResetEvent( port->overlapped.hEvent );
					break;

				case WAIT_TIMEOUT:
					fprintf( stderr, "ser_read() - WaitForSingleObject() timed out\n" );
					break;
				}
			}
		}

	return (int)dwBytesRead;
#endif	// _WIN32

	return 0;
	}


int ser_write(  _ser_struct *port,
                unsigned char *buffer,
                int count,
                int timeout )
	{
	if ( port == NULL )
		{
		fprintf( stdout, "ser_write() - param is NULL!\n" );
		return 0;
		}

#ifdef _WIN32
	DWORD dwBytesWritten = 0;

	DWORD startTime = ::GetTickCount();

	port->overlapped.Offset = 0;
	port->overlapped.OffsetHigh = 0;

	COMSTAT comstat;
	DWORD dwErrorMask;
	::ClearCommError( port->handle, &dwErrorMask, &comstat );

	if ( !::WriteFile( port->handle, buffer, count, &dwBytesWritten, &port->overlapped ) )
		{
		DWORD error = ::GetLastError();
		if ( error != ERROR_IO_PENDING )
			{
			ShowError( "ser_write() - WriteFile()" );
			return 0;
			}
		else
			{
			switch ( ::WaitForSingleObject( port->overlapped.hEvent, timeout ) )
				{
				case WAIT_OBJECT_0:
					if ( ! ::GetOverlappedResult( port->handle, &port->overlapped, &dwBytesWritten, FALSE ) )
						{
						ShowError( "ser_write() (writing data) - GetOverlappedResult()" );
						return 0;
						}
					::ResetEvent( port->overlapped.hEvent );
					break;

				case WAIT_TIMEOUT:
					fprintf( stdout,
							"ser_write() (writing data) - WaitForSingleObject() timed out\n" );
					return 0;
					break;
				}
			}
		}
	else
		{
		}

	DWORD now = ::GetTickCount();
	DWORD elapsed = now - startTime;
	DWORD totalTime = (DWORD)((double)count * (double)port->byteTime);
	DWORD remaining = totalTime - elapsed;
	if ( remaining < totalTime )
		{
		::Sleep( remaining );
		}

	return (int)dwBytesWritten;
#endif	// _WIN32


	return 0;
	}
