#ifndef _SERIAL_H
#define _SERIAL_H

// first, bring in any OS-dependent headers.

#ifdef _WIN32
// For Win32, we'll just get <windows.h>
#include <windows.h>
#endif

#ifdef _WIN32
typedef struct
	{
	HANDLE		handle;
	OVERLAPPED	overlapped;
	double		byteTime;
	} _ser_struct;
#endif // _WIN32

typedef enum
	{
	_ser_stop_1,
	_ser_stop_1_5,
	_ser_stop_2,
	} _eSer_Stop;

typedef enum
	{
	_ser_parity_none,
	_ser_parity_even,
	_ser_parity_odd
	} _eSer_Parity;

////////////////////////////////////////////////////////////////////////
// ser_open( port, bps, data, stop, parity );
//
// Function:    Opens the specified serial port, and configures it
//              with the passed bit rate, data bit count, stop bit
//              count, and parity 
//
// Input:
//    int port
//        A numeric index that represents the serial port to open. In
//        Win32, 0 == COM1, 1 == COM2, etc.
//    int bps
//        Data rate, in bits per second (e.g. 9600, 115200)
//    int data
//        data bits per async word
//    _eSer_Stop stop
//        enum value indicating stop bits per async word
//    _eSer_Parity parity
//        enum value indicating parity type, if any
//
// Output:
//    On success, a pointer to a _ser_struct value representing the
//        opened port
//    On failure, NULL
_ser_struct *ser_open(  int port,
                    int bps,
                    int data,
                    _eSer_Stop stop,
                    _eSer_Parity parity );

////////////////////////////////////////////////////////////////////////
// ser_close( handle );
//
// Function:    Closes the specified port. 
//
// Input:
//    _ser_struct *port
//        The value returned from a successful ser_open() call. The
//        only checking performed on this value is to make sure it's
//        not NULL.
//
// Output:
//    none
void ser_close( _ser_struct *port );

////////////////////////////////////////////////////////////////////////
// ser_configure( handle, bps, data, stop, parity );
//
// Function:    Configures the passed port handle with the specified
//              bit rate, data bit count, stop bit count, and parity 
//
// Input:
//    _ser_struct *port
//        The value returned from a successful ser_open() call. The
//        only checking performed on this value is to make sure it's
//        not NULL.
//    int bps
//        Data rate, in bits per second (e.g. 9600, 115200)
//    int data
//        data bits per async word
//    _eSer_Stop stop
//        enum value indicating stop bits per async word
//    _eSer_Parity parity
//        enum value indicating parity type, if any
//
// Output:
//    On success, non-zero
//    On failure, zero
int ser_configure(  _ser_struct *port,
                    int bps,
                    int data,
                    _eSer_Stop stop,
                    _eSer_Parity parity );


////////////////////////////////////////////////////////////////////////
// ser_peek( port );
//
// Function:    Gets the number of bytes available for reading from
//              the serial port. ser_read() is guaranteed to return
//              immediately if asked to read this number of bytes or
//              less
//
// Input:
//    _ser_struct *port
//        The value returned from a successful ser_open() call. The
//        only checking performed on this value is to make sure it's
//        not NULL.
//
// Output:
//    On success, number of bytes available for read (may be zero)
//    On failure (bad parameter), zero
int ser_peek( _ser_struct *port );


////////////////////////////////////////////////////////////////////////
// ser_read( port, buffer, count, timeout );
//
// Function:    Reads data from the port 
//
// Input:
//    _ser_struct *port
//        The value returned from a successful ser_open() call. The
//        only checking performed on this value is to make sure it's
//        not NULL.
//    unsigned char *buffer
//        Location into which we read the data. This must point to a
//        buffer large enough to hold <count> bytes.
//    int count
//        number of bytes to read
//    int timeout
//        specifies the number of milliseconds to wait if <count> bytes
//        are not immediately available
//
// Output:
//    On success, number of bytes read (may be zero)
//    On failure, zero
int ser_read(   _ser_struct *port,
                unsigned char *buffer,
                int count,
                int timeout );


////////////////////////////////////////////////////////////////////////
// ser_write( port, buffer, count, timeout );
//
// Function:    Writes data to the port 
//
// Input:
//    _ser_struct *port
//        The value returned from a successful ser_open() call. The
//        only checking performed on this value is to make sure it's
//        not NULL.
//    unsigned char *buffer
//        Location from which we write the data.
//    int count
//        number of bytes to write
//    int timeout
//        specifies the number of milliseconds to wait if <count> bytes
//        are not immediately written
//
// Output:
//    On success, number of bytes written (may be zero)
//    On failure, zero
int ser_write(  _ser_struct *port,
                unsigned char *buffer,
                int count,
                int timeout );

#endif
